/* gunicodeprivate.h
 *
 * Copyright (C) 2012 Red Hat, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __G_CHARSET_PRIVATE_H__
#define __G_CHARSET_PRIVATE_H__

#include "gcharset.h"

G_BEGIN_DECLS

const char ** _g_charset_get_aliases (const char *canonical_name);

gboolean      _g_get_time_charset    (const char **charset);

gboolean      _g_get_ctype_charset   (const char **charset);

G_END_DECLS

#endif
