/* GIO - GLib Input, Output and Streaming Library
 *
 * Copyright 2023 Leo Zi-You Assini <leoziyou@amazon.it>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __G_OSX_NETWORK_MONITOR_H__
#define __G_OSX_NETWORK_MONITOR_H__

#include "gnetworkmonitorbase.h"

G_BEGIN_DECLS

#define G_TYPE_OSX_NETWORK_MONITOR (g_osx_network_monitor_get_type ())
G_DECLARE_FINAL_TYPE (GOsxNetworkMonitor, g_osx_network_monitor, G, OSX_NETWORK_MONITOR, GNetworkMonitorBase);

G_END_DECLS

#endif /* __G_OSX_NETWORK_MONITOR_H__ */